package de.fhdw.gaming.memory.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.List;

import de.fhdw.gaming.core.domain.Game;
import de.fhdw.gaming.core.domain.Move;
import de.fhdw.gaming.core.domain.Player;
import de.fhdw.gaming.core.domain.Strategy;
import de.fhdw.gaming.memory.Brain;
import de.fhdw.gaming.memory.Key;
import de.fhdw.gaming.memory.ShiftList;


/**
 * A class managing all memories of all players.
 */
public final class MemoryBrainImpl implements Brain {
    
    /**
     * Returns the single instance of the MemoryBrainImpl.
     */
    private static Brain instance = new MemoryBrainImpl();
    
    /**
     * The number of outcomes that can be remembered. -1 for as much as possible.
     */
    private int memoryLength = 0;
    
    /**
     * A Map linking a key to a corresponding ShiftList.
     */
    private Map<Key, ShiftList> memory = new HashMap<>(); 
    
    /**
     * A private constructor to prevent it being accessed from outside.
     */
    private MemoryBrainImpl() {
    }
    
    /**
     * Returns the singleton instance of MemoryBrainImpl.
     */
    public static Brain getInstance() {
        return instance;
    }
    
    /**
     * .
     * @return
     */
    public int getLength() {
        return memoryLength;
    }
    
    
    /**
     * Sets the amount of outcomes that the brain can remember. Can be used during a game.
     * @param length
     */
    @Override
    public void setMemoryLength(int length) throws IllegalArgumentException {
        if (length >= -1) {
            memoryLength = length;
            for (Map.Entry<Key, ShiftList> entry : memory.entrySet()) {
                final ShiftList memoryShiftList = entry.getValue();
                memoryShiftList.setLength(memoryLength);
            }
        } else {
            throw new IllegalArgumentException();
        }
    }
    
    /**
     * Adds the outcomes of the current game for each player to the memory. Each combination of active player,
     * their opponent, and the strategy of the active player requires a new {@link Key} for which the outcomes
     * will be saved. In a 2 player game without changing strategies there should
     * only ever be two {@link Key}s.
     * @param game
     */
    @Override
    public void rememberOutcome(Game<?, ?, ?, ?> game) {
        
        final Map<String, ? extends Strategy<?, ?, ?>> tempStrategies = game.getStrategies();
        final Map<String, ? extends Player<?>> tempPlayers = game.getPlayers();
        final List<String> playerNames = new ArrayList<>(tempPlayers.keySet());
        final Set<Key> memoryKeys = memory.keySet();
        
        final Key currentKey1 = new MemoryKeyImpl(tempPlayers
                .get(playerNames.get(0)), tempPlayers.get(playerNames.get(1)), tempStrategies.get(playerNames.get(0)));
        
        final Key currentKey2 = new MemoryKeyImpl(tempPlayers
                .get(playerNames.get(1)), tempPlayers.get(playerNames.get(0)), tempStrategies.get(playerNames.get(1)));
        
        if (memoryKeys.contains(currentKey1)) {
            memory.get(currentKey1).push(currentKey1.getPlayer().getOutcome());
        } else {
            final MemoryShiftListImpl newShiftList1 = new MemoryShiftListImpl(memoryLength);
            newShiftList1.push(currentKey1.getPlayer().getOutcome());
            memory.put(currentKey1, newShiftList1);
        }
        if (memoryKeys.contains(currentKey2)) {
            memory.get(currentKey2).push(currentKey2.getPlayer().getOutcome());
        } else {
            final MemoryShiftListImpl newShiftList2 = new MemoryShiftListImpl(memoryLength);
            newShiftList2.push(currentKey2.getPlayer().getOutcome());
            memory.put(currentKey2, newShiftList2);
        }
        
        this.setMemoryLength(memoryLength);
        
    }
    
    /**
     * .
     * @param key
     * @param move
     */
    public void rememberMyMove(Key key, Move<?, ?> move) {
        if (memory.keySet().contains(key)) {
            getMemory(key).pushMyMove(move);
        } else {
            final MemoryShiftListImpl newShiftList1 = new MemoryShiftListImpl(memoryLength);
            
            newShiftList1.pushMyMove(move);
            
            memory.put(key, newShiftList1);
        }
        
        this.setMemoryLength(memoryLength);
    }
    
    /**
     * .
     * @param key
     * @param move
     */
    public void rememberOpponentMove(Key key, Move<?, ?> move) {
        if (memory.keySet().contains(key)) {
            getMemory(key).pushOpponentMove(move);
        } else {
            final MemoryShiftListImpl newShiftList1 = new MemoryShiftListImpl(memoryLength);
            
            newShiftList1.pushOpponentMove(move);
            
            memory.put(key, newShiftList1);
        }
        
        this.setMemoryLength(memoryLength);
    }
    
    /**
     * Returns the {@link ShiftList} containing the outcomes for a certain {@link Key}.
     * That means a certain combination of player the memory is for,
     * their opponent and the strategy used by the active player.
     * If the {@link Key} is not in memory, returns null.
     * @param key {@link Key}
     */
    @Override
    public ShiftList getMemory(Key key) {
        final Set<Key> memoryKeys = memory.keySet();
        if (memoryKeys.contains(key)) {
            return memory.get(key);
        } else {
            return null;
        }
    }
    
    /**
     * Returns the entirety of the memory. It can be used to acquire any
     * {@link ShiftList} of outcomes with the right key.
     */
    @Override
    public Map<Key, ShiftList> getMemory() {
        return memory;
    }
    
    /**
     * Resets the brain.
     */
    public static void strongHitOnTheHead() {
        instance = new MemoryBrainImpl();
    }
    
    
}
